import java.util.ArrayList;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.SortedSet;
import java.util.TreeSet;

public class Dijkstra {
	public static final double INFINITE_DISTANCE = Double.MAX_VALUE;
    
    private final Comparator shortestDistanceComparator = new Comparator(){
    	 public int compare(Object left, Object right){
             double shortestDistanceLeft = getShortestDistance((MyPoint) left);
             double shortestDistanceRight = getShortestDistance((MyPoint) right);
             
             if (shortestDistanceLeft > shortestDistanceRight){
                 return +1;
             }
             else if (shortestDistanceLeft < shortestDistanceRight){
                 return -1;
             }
             else{
            	 return ((MyPoint) left).compareTo(right);
             }
         }
    };
    
    //graphe reseau
    private final Reseau res;
    
    //liste de noeuds non traits
    //la liste est trie grace au comparateur ci dessus
    private final SortedSet unsettledNodes = new TreeSet(shortestDistanceComparator);
    
    //liste des noeuds deja traits
    private final Set settledNodes = new HashSet();
    
    //liste de plus courtes distances entre chaque noeud et le depart
    private final Map shortestDistances = new HashMap();

    //liste des prdcesseurs de chaque noeud
    private final Map predecessors = new HashMap();
    
    private Trajet trajet=null;
    
    public Dijkstra(Reseau _res){
        res = _res;
        trajet = new Trajet();
    }

    //initialisation de l'algorithme
    private void init(MyPoint start){
        settledNodes.clear();
        unsettledNodes.clear();
        
        shortestDistances.clear();
        predecessors.clear();
        
        // add source
        setShortestDistance(start, 0);
        unsettledNodes.add(start);
    }
    
    
    //execution de l'algorithme de Dijkstra sur le reseau
    public void execute(MyPoint start, MyPoint destination){
        init(start);
        // the current node
        MyPoint u;
        // extract the node with the shortest distance
        while((u = extractMin()) != null){
            assert !isSettled(u);
            // destination reached, stop
            if (u.getNum() == destination.getNum()) break;
            
            markSettled(u);
            
            relaxNeighbors(u);
        }
        
        makeTrajet(start, destination);
    }

    //retourne le noeud ayant la distance depuis le dpart la plus faible et n'ayant pas t trait
    private MyPoint extractMin(){
    	if (unsettledNodes.isEmpty()) return null;
    	
        MyPoint min = (MyPoint) unsettledNodes.first();
        unsettledNodes.remove(min);
        
        return min;
    }
    
	//mise  jour des distances aux noeuds suivants
    private void relaxNeighbors(MyPoint u){
        for(Iterator i = res.getDestinations(u).iterator(); i.hasNext();){
            MyPoint v = (MyPoint) i.next();
           
            //ne pas traiter les noeuds dja traits
            if (!isSettled(v)){            
	            if (getShortestDistance(v) > getShortestDistance(u) + res.getDistance(u, v)){
	            	//assigner la nouvelle distance la plus courte et mise  jour des noeuds non traits
	                setShortestDistance(v, getShortestDistance(u) + res.getDistance(u, v));
	
	                //assigner un prdecesseur au noeud
	                setPredecessor(v, u);
	            }
            }
        }        
    }

	//ajouter un noeud dans la liste des noeuds traits
	private void markSettled(MyPoint u){
		settledNodes.add(u);  
	}

	//test si un noeud a dja t trait
    private boolean isSettled(MyPoint v){
        return settledNodes.contains(v);
    }

    //retourne la plus courte distance entre le noeud et le dpart 
    public double getShortestDistance(MyPoint mp){
        Double d = (Double)shortestDistances.get(mp);
        return (d == null) ? INFINITE_DISTANCE : d.doubleValue();
    }

	//mettre  jour la nouvelle distance la plus courte     
    private void setShortestDistance(MyPoint mp, double distance){
        // this crucial step ensure no duplicates will be created in the queue
        // when an existing unsettled node is updated with a new shortest distance

        unsettledNodes.remove((Object)mp);

        shortestDistances.put(mp, new Double(distance));
        
		// re-balance the sorted set according to the new shortest distance found
		// (see the comparator the set was initialized with)
		unsettledNodes.add(mp);
	}
    
    //retourne le prdecesseur d'un noeud en suivant le plus court chemin
    public MyPoint getPredecessor(MyPoint mp){
        return (MyPoint)predecessors.get(mp);
    }
    
    //ajouter un noeud dans la liste des prdecesseurs
    private void setPredecessor(MyPoint a, MyPoint b){
        predecessors.put(a, b);
    }
    
    //cration du trajet de plus courte distance entre les points start et destination
    private void makeTrajet(MyPoint start, MyPoint destination){
    	MyPoint pt=destination;
    	
		while(getPredecessor(pt).getNum()!=start.getNum()){
			trajet.addArc(new Arc(getPredecessor(pt), pt));
			pt=getPredecessor(pt);
		}
		trajet.addArc(new Arc(pt,start));
		
		ArrayList<Rue> lr=res.getListRues();
    	
		for(int i=0;i<trajet.getListArcs().size();i++){
			for(int j=0;j<lr.size();j++){
				if(lr.get(j).containsArc(trajet.getListArcs().get(i))){
					trajet.addRue(lr.get(j));
				}
			}
		}
    }
    
    public Trajet getTrajet(){
    	return trajet;
    }
    
}
